class SegmentQueueElement {

    public Segment seg;
    public SegmentQueueElement next;

    public SegmentQueueElement(Segment s) {
	seg = s;
        next = null;
    }

}

public class Train extends RailwayNetwork implements Runnable {

    String trainId;
    SegmentQueueElement trainEndLocation;
    SegmentQueueElement trainHeadLocation;
    SegmentQueueElement initialLocation;
    int     trainDirection;
    int     trainSpeed;

    public Train(String tid, Segment loc, int dir) {

        trainId = tid;
        trainEndLocation = new SegmentQueueElement(loc);
        initialLocation = trainHeadLocation = trainEndLocation;
        trainSpeed = STOPPED;
        trainDirection = dir;
        try { trainHeadLocation.seg.enterSegment(this); }
        catch ( Exception e ) { e.printStackTrace(); }
    }

    void reinitialise() {
	trainHeadLocation = trainEndLocation = initialLocation;
        initialLocation.next = null;
        trainSpeed = STOPPED;
        try { trainHeadLocation.seg.enterSegment(this); }
        catch ( Exception e ) { e.printStackTrace(); }
    }

    public String toString() { return trainId; }

    public boolean isGone() { return trainHeadLocation == null; }

    public int getDirection() { return trainDirection; }

    public int getSpeed() { return trainSpeed; }

    public void move() {

	switch(trainSpeed) {
	    case STOPPED: if ( Math.random() < 0.5 ) 
		trainSpeed = DRIVING; 
	    System.out.println(toString() 
			       + ": DRIVING from "
                               + trainHeadLocation.seg.toString()
                               + "\n");
	    return;
	    case EMERGENCY_STOPPED: return;
	    case DRIVING: moveOneStep();
		if ( isGone() ) return;
		System.out.println(toString() 
				   + ": DRIVING in "
				   + ((trainHeadLocation.seg != null) 
				      ? trainHeadLocation.seg.toString()
				      : " ")
				   + "<--"
				   + ((trainEndLocation.seg != null)
				      ? trainEndLocation.seg.toString()
				      : " ")
				   + "\n");
		return;
	    case BRAKING: moveOneStep(); 
		trainSpeed = STOPPED; 
		return;
	}

    }

    public void run() {

       while ( true ) {
	    try { Thread.sleep((int)(2000*Math.random())); }
            catch (InterruptedException e) { e.printStackTrace(); }
            move();
            if (isGone()) reinitialise();
	}

    }

    /////////////// non-public methods /////////////////////////////
    void moveOneStep() {

        // this train no longer exists if trainHeadLocation == null
        if (trainHeadLocation == null) return;

	switch(trainDirection) {
	    case NO_DIR: return;
	    case LEFT:
	    case RIGHT:
		do {
		    // are we leaving the network 
                    // under consideration ?
		    if ( trainHeadLocation.seg.getNeighbour
			 (trainDirection) 
			 == null ) {
			// leave the segment at the train end
			trainEndLocation.seg.leaveSegment(this);
			trainEndLocation = trainEndLocation.next;
			// has the train disappeared ?
			if ( trainEndLocation == null ) 
			    trainHeadLocation = null;
		    }
		    // for all remaining cases,
                    // we are still inside the network
		    else if ( trainEndLocation == trainHeadLocation ) {
                        // we are completly inside one segment:
                        // now we enter the next segment, but 
                        // still stay (partially) in the last
                        // segment.
                        trainHeadLocation.next 
			= new SegmentQueueElement
			    (trainHeadLocation.seg.getNeighbour(trainDirection));
                        trainHeadLocation = trainHeadLocation.next;
			try { trainHeadLocation.seg.enterSegment(this); }
			catch ( Exception e ) { e.printStackTrace(); }
		    }
                    else if ( trainHeadLocation != trainEndLocation
                              && ! trainHeadLocation.seg.isShort() ) {
			// Here the head of the train is located on a long segment.
                        // The reset of the train resides on one (long) or more (short)
                        // segments.
                        // The train head is in a long segment, and remains there
                        // while the train-end-segment is left,
                        trainEndLocation.seg.leaveSegment(this);
			trainEndLocation = trainEndLocation.next;
		    } 
		    else if ( trainHeadLocation.seg.isShort() ) {
			// here the train head resides in a short segment
                        // we move the head immediately to the next segment
                        trainHeadLocation.next 
			= new SegmentQueueElement
			    (trainHeadLocation.seg.getNeighbour(trainDirection));
                        trainHeadLocation = trainHeadLocation.next;
			try { trainHeadLocation.seg.enterSegment(this); }
			catch ( Exception e ) { e.printStackTrace(); }
		    }

	} while (trainHeadLocation != null 
		 && trainHeadLocation.seg.isShort());
		return;
	}
     
    }

}






